"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ParseMessage = exports.FullMessage = void 0;
const zlib_1 = __importDefault(require("zlib"));
const util_1 = require("util");
const bson_1 = require("bson");
class MessageHeader {
    constructor(values) {
        this.messageLength = values.messageLength;
        this.requestID = values.requestID;
        this.responseTo = values.responseTo;
        this.opCode = values.opCode;
    }
    static async Read(buf) {
        const dv = new DataView(buf.buffer, buf.byteOffset, 4 * 4);
        return {
            msg: new this({
                messageLength: dv.getInt32(0, true),
                requestID: dv.getInt32(4, true),
                responseTo: dv.getInt32(8, true),
                opCode: dv.getInt32(12, true)
            }),
            readBytes: 4 * 4
        };
    }
}
class OpCode {
    static LookupOpCode(value) {
        for (const key of Object.keys(OpCode)) {
            if (OpCode[key] === value) {
                return key;
            }
        }
        return null;
    }
}
OpCode.OP_MSG = 2013;
OpCode.OP_REPLY = 1;
OpCode.OP_UPDATE = 2001;
OpCode.OP_INSERT = 2002;
OpCode.OP_QUERY = 2004;
OpCode.OP_GET_MORE = 2005;
OpCode.OP_DELETE = 2006;
OpCode.OP_KILL_CURSORS = 2007;
OpCode.OP_COMPRESSED = 2012;
class BSONBuffer {
    constructor(buf) {
        this.data = bson_1.deserialize(buf);
    }
}
class OpMsgBodySection {
    constructor(body) {
        this.kind = 'Body';
        this.body = body;
    }
}
class OpMsgDocumentSequenceSection {
    constructor(docSequenceId, objects) {
        this.kind = 'DocumentSequence';
        this.docSequenceId = docSequenceId;
        this.objects = objects;
    }
}
class OpMsgUnknownSection {
    constructor(kindByte) {
        this.kind = 'Unknown';
        this.kindByte = kindByte;
    }
}
class OpMsg {
    constructor(values) {
        this.opCode = 'OP_MSG';
        this.flagBits = values.flagBits;
        this.sections = values.sections;
        this.checksum = values.checksum;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        const flagBits = dv.getUint32(0, true);
        const checksum = (flagBits & 1) ? dv.getUint32(length - 4, true) : null;
        const sections = [];
        let i;
        for (i = 4; i < ((flagBits & 1) ? length - 4 : length);) {
            const kind = dv.getUint8(i);
            switch (kind) {
                case 0:
                    {
                        const bsonSize = dv.getUint32(i + 1, true);
                        sections.push(new OpMsgBodySection(new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i + 1, bsonSize))));
                        i += 1 + bsonSize;
                    }
                    break;
                case 1:
                    {
                        const sectionSize = dv.getUint32(i + 1, true);
                        const seqIdStart = i + 5;
                        for (i = seqIdStart; i < seqIdStart - 4 + sectionSize; i++) {
                            if (dv.getUint8(i) === 0) {
                                break;
                            }
                        }
                        const docSequenceId = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + seqIdStart, i - 1 - seqIdStart));
                        const objects = [];
                        for (i = i + 1; i < seqIdStart - 4 + sectionSize;) {
                            const bsonSize = dv.getUint32(i, true);
                            objects.push(new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize)));
                            i += bsonSize;
                        }
                        sections.push(new OpMsgDocumentSequenceSection(docSequenceId, objects));
                    }
                    break;
                default: {
                    sections.push(new OpMsgUnknownSection(kind));
                    i = length;
                    break;
                }
            }
        }
        return {
            msg: new this({
                flagBits,
                sections,
                checksum
            }),
            readBytes: i + (flagBits & 1) ? 4 : 0
        };
    }
}
class OpUpdate {
    constructor(values) {
        this.opCode = 'OP_UPDATE';
        this.fullCollectionName = values.fullCollectionName;
        this.flags = values.flags;
        this.selector = values.selector;
        this.update = values.update;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i;
        for (i = 4; i < length; i++) {
            if (dv.getUint8(i) === 0) {
                break;
            }
        }
        const fullCollectionName = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + 4, i - 4));
        i++;
        const flags = dv.getInt32(i, true);
        let selector;
        let update;
        {
            const bsonSize = dv.getUint32(i, true);
            selector = new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize));
            i += bsonSize;
        }
        {
            const bsonSize = dv.getUint32(i, true);
            update = new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize));
            i += bsonSize;
        }
        return {
            msg: new this({
                fullCollectionName,
                flags,
                selector,
                update
            }),
            readBytes: i
        };
    }
}
class OpInsert {
    constructor(values) {
        this.opCode = 'OP_INSERT';
        this.fullCollectionName = values.fullCollectionName;
        this.flags = values.flags;
        this.documents = values.documents;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i = 0;
        const flags = dv.getInt32(i, true);
        for (i = 4; i < length; i++) {
            if (dv.getUint8(i) === 0) {
                break;
            }
        }
        const fullCollectionName = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + 4, i - 4));
        i++;
        const documents = [];
        while (i < length) {
            const bsonSize = dv.getUint32(i, true);
            documents.push(new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize)));
            i += bsonSize;
        }
        return {
            msg: new this({
                fullCollectionName,
                flags,
                documents
            }),
            readBytes: i
        };
    }
}
class OpQuery {
    constructor(values) {
        this.opCode = 'OP_QUERY';
        this.fullCollectionName = values.fullCollectionName;
        this.flags = values.flags;
        this.numberToSkip = values.numberToSkip;
        this.numberToReturn = values.numberToReturn;
        this.query = values.query;
        this.returnFieldsSelector = values.returnFieldsSelector;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i = 0;
        const flags = dv.getInt32(i, true);
        for (i = 4; i < length; i++) {
            if (dv.getUint8(i) === 0) {
                break;
            }
        }
        const fullCollectionName = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + 4, i - 4));
        i++;
        const numberToSkip = dv.getInt32(i, true);
        i += 4;
        const numberToReturn = dv.getInt32(i, true);
        i += 4;
        let query;
        let returnFieldsSelector = null;
        {
            const bsonSize = dv.getUint32(i, true);
            query = new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize));
            i += bsonSize;
        }
        if (i < length) {
            const bsonSize = dv.getUint32(i, true);
            returnFieldsSelector = new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize));
            i += bsonSize;
        }
        return {
            msg: new this({
                fullCollectionName,
                flags,
                numberToSkip,
                numberToReturn,
                query,
                returnFieldsSelector
            }),
            readBytes: i
        };
    }
}
class OpGetMore {
    constructor(values) {
        this.opCode = 'OP_GET_MORE';
        this.fullCollectionName = values.fullCollectionName;
        this.numberToReturn = values.numberToReturn;
        this.cursorId = values.cursorId;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i;
        for (i = 4; i < length; i++) {
            if (dv.getUint8(i) === 0) {
                break;
            }
        }
        const fullCollectionName = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + 4, i - 4));
        i++;
        const numberToReturn = dv.getInt32(i, true);
        i += 4;
        const cursorId = String(dv.getBigInt64(i, true));
        i += 8;
        return {
            msg: new this({
                fullCollectionName,
                numberToReturn,
                cursorId
            }),
            readBytes: i
        };
    }
}
class OpDelete {
    constructor(values) {
        this.opCode = 'OP_DELETE';
        this.fullCollectionName = values.fullCollectionName;
        this.flags = values.flags;
        this.selector = values.selector;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i;
        for (i = 4; i < length; i++) {
            if (dv.getUint8(i) === 0) {
                break;
            }
        }
        const fullCollectionName = new util_1.TextDecoder().decode(new Uint8Array(dv.buffer, dv.byteOffset + 4, i - 4));
        i++;
        const flags = dv.getInt32(i, true);
        i += 4;
        let selector;
        {
            const bsonSize = dv.getUint32(i, true);
            selector = new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize));
            i += bsonSize;
        }
        return {
            msg: new this({
                fullCollectionName,
                flags,
                selector
            }),
            readBytes: i
        };
    }
}
class OpKillCursors {
    constructor(values) {
        this.opCode = 'OP_KILL_CURSORS';
        this.cursorIds = values.cursorIds;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        const n = dv.getInt32(4, true);
        const cursorIds = [];
        let i;
        for (i = 8; (i - 8) < n * 8; i += 8) {
            cursorIds.push(String(dv.getBigInt64(i, true)));
        }
        return {
            msg: new this({
                cursorIds
            }),
            readBytes: i
        };
    }
}
class OpCompressed {
    constructor(values) {
        this.opCode = 'OP_COMPRESSED';
        this.originalOpcode = values.originalOpcode;
        this.compressorId = values.compressorId;
        this.contents = values.contents;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        const originalOpcode = dv.getInt32(0, true);
        const compressorIdN = dv.getUint8(8);
        let compressorId;
        switch (compressorIdN) {
            case 0:
                compressorId = 'noop';
                break;
            case 1:
                compressorId = 'snappy';
                break;
            case 2:
                compressorId = 'zlib';
                break;
            case 3:
                compressorId = 'zstd';
                break;
            default:
                compressorId = `unknown (${compressorIdN})`;
                break;
        }
        const compressed = new Uint8Array(dv.buffer, dv.byteOffset + 9, length - 9);
        let uncompressed = null;
        if (compressorId === 'noop') {
            uncompressed = compressed;
        }
        else if (compressorId === 'zlib') {
            uncompressed = await util_1.promisify(zlib_1.default.inflate)(compressed);
        }
        let contents;
        if (uncompressed === null) {
            contents = compressed;
        }
        else {
            contents = (await ReadOpContents(originalOpcode, uncompressed, uncompressed.length)).msg;
        }
        return {
            msg: new this({
                originalOpcode,
                compressorId,
                contents
            }),
            readBytes: length
        };
    }
}
class OpReply {
    constructor(values) {
        this.opCode = 'OP_REPLY';
        this.flags = values.flags;
        this.cursorId = values.cursorId;
        this.startingFrom = values.startingFrom;
        this.numberReturned = values.numberReturned;
        this.documents = values.documents;
    }
    static async Read(buf, length) {
        const dv = new DataView(buf.buffer, buf.byteOffset, length);
        let i = 0;
        const flags = dv.getInt32(i, true);
        i += 4;
        const cursorId = String(dv.getBigUint64(i, true));
        i += 8;
        const startingFrom = dv.getInt32(i, true);
        i += 4;
        const numberReturned = dv.getInt32(i, true);
        i += 4;
        const documents = [];
        while (i < length) {
            const bsonSize = dv.getUint32(i, true);
            documents.push(new BSONBuffer(new Uint8Array(dv.buffer, dv.byteOffset + i, bsonSize)));
            i += bsonSize;
        }
        return {
            msg: new this({
                flags,
                cursorId,
                startingFrom,
                numberReturned,
                documents
            }),
            readBytes: i
        };
    }
}
class OpUnknown {
    constructor(values) {
        this.opCode = 'OP_UNKNOWN';
        this.actualOpCode = values.actualOpCode;
        this.body = values.body;
    }
}
async function ReadOpContents(opcode, data, length) {
    switch (opcode) {
        case OpCode.OP_MSG: return OpMsg.Read(data, length);
        case OpCode.OP_REPLY: return OpReply.Read(data, length);
        case OpCode.OP_UPDATE: return OpUpdate.Read(data, length);
        case OpCode.OP_INSERT: return OpInsert.Read(data, length);
        case OpCode.OP_QUERY: return OpQuery.Read(data, length);
        case OpCode.OP_GET_MORE: return OpGetMore.Read(data, length);
        case OpCode.OP_DELETE: return OpDelete.Read(data, length);
        case OpCode.OP_KILL_CURSORS: return OpKillCursors.Read(data, length);
        case OpCode.OP_COMPRESSED: return OpCompressed.Read(data, length);
        default: {
            return {
                msg: new OpUnknown({ actualOpCode: opcode, body: data.subarray(0, length) }),
                readBytes: length
            };
        }
    }
}
class FullMessage {
    constructor(header, contents) {
        this.header = header;
        this.contents = contents;
    }
}
exports.FullMessage = FullMessage;
async function ParseMessage(data) {
    if (data.length < 16) {
        return { needBytes: 16 };
    }
    const { msg: header, readBytes: headerBytes } = await MessageHeader.Read(data);
    if (header.messageLength > data.length) {
        return { needBytes: data.length - header.messageLength };
    }
    const contents = (await ReadOpContents(header.opCode, data.subarray(headerBytes, header.messageLength), header.messageLength - headerBytes)).msg;
    return {
        msg: new FullMessage(header, contents),
        readBytes: header.messageLength
    };
}
exports.ParseMessage = ParseMessage;
//# sourceMappingURL=parse.js.map